from __future__ import annotations

from asphalt.core import qualified_name


def default_marshaller(obj):
    """
    Retrieve the state of the given object.

    Calls the ``__getstate__()`` method of the object if available, otherwise returns the
    ``__dict__`` of the object.

    :param obj: the object to marshal
    :return: the marshalled object state

    """
    if hasattr(obj, '__getstate__'):
        return obj.__getstate__()

    try:
        return obj.__dict__
    except AttributeError:
        raise TypeError('{!r} has no __dict__ attribute and does not implement __getstate__()'
                        .format(qualified_name(obj.__class__))) from None


def default_unmarshaller(instance, state) -> None:
    """
    Restore the state of an object.

    If the ``__setstate__()`` method exists on the instance, it is called with the state object
    as the argument. Otherwise, the instance's ``__dict__`` is replaced with ``state``.

    :param instance: an uninitialized instance
    :param state: the state object, as returned by :func:`default_marshaller`

    """
    if hasattr(instance, '__setstate__'):
        instance.__setstate__(state)
    else:
        try:
            instance.__dict__.update(state)
        except AttributeError:
            raise TypeError('{!r} has no __dict__ attribute and does not implement __setstate__()'
                            .format(qualified_name(instance.__class__))) from None



import pytest
from asphalt.core.context import Context

from asphalt.serialization.api import Serializer
from asphalt.serialization.component import SerializationComponent
from asphalt.serialization.serializers.json import JSONSerializer


@pytest.mark.asyncio
async def test_component_start():
    component = SerializationComponent(serializers={
        'json': {'encoder_options': {'allow_nan': False}},
        'msgpack': {'unpacker_options': {'encoding': 'iso-8859-1'}},
        'pickle': {'protocol': 3},
        'yaml': {'safe': False}
    })
    async with Context() as ctx:
        await component.start(ctx)

        assert ctx.json
        assert ctx.msgpack
        assert ctx.pickle
        assert ctx.yaml


@pytest.mark.asyncio
async def test_default_config():
    component = SerializationComponent(backend='json')
    async with Context() as ctx:
        await component.start(ctx)

        resource = ctx.require_resource(Serializer)
        assert isinstance(resource, JSONSerializer)
        assert ctx.json is resource


@pytest.mark.asyncio
async def test_null_configs():
    component = SerializationComponent(serializers={
        'json': None,
        'msgpack': None,
        'pickle': None,
        'yaml': None
    })
    async with Context() as ctx:
        await component.start(ctx)

        assert ctx.json
        assert ctx.msgpack
        assert ctx.pickle
        assert ctx.yaml



from __future__ import annotations

from io import StringIO

from ruamel.yaml import YAML
from typeguard import check_argument_types

from asphalt.serialization.api import Serializer


class YAMLSerializer(Serializer):
    """
    Serializes objects to and from YAML format.

    To use this serializer backend, the ``ruamel.yaml`` library must be installed.
    A convenient way to do this is to install ``asphalt-serialization`` with the ``yaml``
    extra:

    .. code-block:: shell

        $ pip install asphalt-serialization[yaml]

    .. warning:: This serializer is insecure in unsafe mode because it allows execution of
      arbitrary code when deserializing.

    .. seealso:: `PyYAML documentation <http://pyyaml.org/wiki/PyYAMLDocumentation>`_

    :param safe: ``True`` to (de)serialize in safe mode, ``False`` to enable extended tags
    """

    __slots__ = '_yaml', '_dumper_options'

    def __init__(self, safe: bool = True):
        assert check_argument_types()
        self._yaml = YAML(typ='safe' if safe else 'unsafe')

    def serialize(self, obj) -> bytes:
        buffer = StringIO()
        self._yaml.dump(obj, buffer)
        return buffer.getvalue().encode('utf-8')

    def deserialize(self, payload: bytes):
        return self._yaml.load(payload)

    @property
    def mimetype(self):
        return 'text/yaml'

    @property
    def safe(self):
        """Returns ``True`` if the safe mode is being used with (de)serialization."""
        return 'safe' in self._yaml.typ



from __future__ import annotations

from json.decoder import JSONDecoder
from json.encoder import JSONEncoder
from typing import Any, Dict, Union

from asphalt.core import resolve_reference
from typeguard import check_argument_types

from asphalt.serialization.api import CustomizableSerializer
from asphalt.serialization.object_codec import DefaultCustomTypeCodec


class JSONTypeCodec(DefaultCustomTypeCodec):
    """Default state wrapper implementation for :class:`~.JSONSerializer`."""

    def register_object_encoder_hook(self, serializer: 'JSONSerializer') -> None:
        self.serializer = serializer
        serializer.encoder_options['default'] = self.default_encoder
        serializer._encoder = JSONEncoder(**serializer.encoder_options)

    def register_object_decoder_hook(self, serializer: 'JSONSerializer') -> None:
        self.serializer = serializer
        serializer.decoder_options['object_hook'] = self.default_decoder
        serializer.decoder_options.pop('object_pairs_hook', None)
        serializer._decoder = JSONDecoder(**serializer.decoder_options)


class JSONSerializer(CustomizableSerializer):
    """
    Serializes objects using JSON (JavaScript Object Notation).

    See the :mod:`json` module documentation in the standard library for more information on
    available options.

    Certain options can resolve references to objects:

    * ``encoder_options['default']``
    * ``decoder_options['object_hook']``
    * ``decoder_options['object_pairs_hook']``

    :param encoder_options: keyword arguments passed to :class:`~json.JSONEncoder`
    :param decoder_options: keyword arguments passed to :class:`~json.JSONDecoder`
    :param encoding: the text encoding to use for converting to and from bytes
    :param custom_type_codec: wrapper to use to wrap custom types after marshalling
    """

    __slots__ = ('encoder_options', 'decoder_options', 'encoding', 'custom_type_codec',
                 '_encoder', '_decoder', '_marshallers', '_unmarshallers')

    def __init__(self, encoder_options: Dict[str, Any] = None,
                 decoder_options: Dict[str, Any] = None, encoding: str = 'utf-8',
                 custom_type_codec: Union[JSONTypeCodec, str] = None):
        assert check_argument_types()
        super().__init__(resolve_reference(custom_type_codec) or JSONTypeCodec())
        self.encoding = encoding

        self.encoder_options = encoder_options or {}
        self.encoder_options['default'] = resolve_reference(self.encoder_options.get('default'))
        self._encoder = JSONEncoder(**self.encoder_options)

        self.decoder_options = decoder_options or {}
        self.decoder_options['object_hook'] = resolve_reference(
            self.decoder_options.get('object_hook'))
        self.decoder_options['object_pairs_hook'] = resolve_reference(
            self.decoder_options.get('object_pairs_hook'))
        self._decoder = JSONDecoder(**self.decoder_options)

    def serialize(self, obj) -> bytes:
        return self._encoder.encode(obj).encode(self.encoding)

    def deserialize(self, payload: bytes):
        text_payload = payload.decode(self.encoding)
        return self._decoder.decode(text_payload)

    @property
    def mimetype(self):
        return 'application/json'



from __future__ import annotations

from typing import Any, Dict, Optional, Union

import cbor2
from asphalt.core import qualified_name, resolve_reference
from typeguard import check_argument_types

from asphalt.serialization.api import CustomizableSerializer
from asphalt.serialization.object_codec import DefaultCustomTypeCodec


class CBORTypeCodec(DefaultCustomTypeCodec):
    """
    Default custom type codec implementation for :class:`~.CBORSerializer`.

    Wraps marshalled state in either CBORTag objects (the default) or dicts (with
    ``type_tag=None``).

    :param type_tag: CBOR tag number to use, or ``None`` to use JSON compatible dict-based
        wrapping

    .. note:: Custom wrapping hooks are ignored when CBORTags are used.
    """

    def __init__(self, type_tag: Optional[int] = 4554, **kwargs):
        super().__init__(**kwargs)
        self.type_tag = type_tag

    def register_object_encoder_hook(self, serializer: 'CBORSerializer') -> None:
        self.serializer = serializer
        if self.type_tag:
            serializer.encoder_options['default'] = cbor2.shareable_encoder(self.cbor_tag_encoder)
        else:
            serializer.encoder_options['default'] = self.cbor_default_encoder

    def register_object_decoder_hook(self, serializer: 'CBORSerializer') -> None:
        self.serializer = serializer
        if self.type_tag:
            serializer.decoder_options['tag_hook'] = self.cbor_tag_decoder
        else:
            serializer.decoder_options['object_hook'] = self.cbor_default_decoder

    def cbor_tag_encoder(self, encoder: cbor2.CBOREncoder, obj):
        try:
            typename, marshaller, wrap_state = self.serializer.marshallers[obj.__class__]
        except KeyError:
            raise LookupError('no marshaller found for type "{}"'
                              .format(qualified_name(type(obj)))) from None

        marshalled_state = marshaller(obj)
        if wrap_state:
            serialized_state = encoder.encode_to_bytes(marshalled_state)
            wrapped_state = [typename, serialized_state]
            encoder.encode(cbor2.CBORTag(self.type_tag, wrapped_state))
        else:
            encoder.encode(marshalled_state)

    def cbor_tag_decoder(self, decoder: cbor2.CBORDecoder, tag: cbor2.CBORTag):
        if tag.tag != self.type_tag:
            return tag

        typename, serialized_state = tag.value
        try:
            cls, unmarshaller = self.serializer.unmarshallers[typename]
        except KeyError:
            raise LookupError(f'no unmarshaller found for type "{typename}"') from None

        if cls is not None:
            instance = cls.__new__(cls)
            decoder.set_shareable(instance)
            marshalled_state = decoder.decode_from_bytes(serialized_state)
            unmarshaller(instance, marshalled_state)
            return instance
        else:
            marshalled_state = decoder.decode_from_bytes(serialized_state)
            return unmarshaller(marshalled_state)

    def cbor_default_encoder(self, encoder: cbor2.CBOREncoder, obj):
        encoded = self.default_encoder(obj)
        encoder.encode(encoded)

    def cbor_default_decoder(self, decoder: cbor2.CBORDecoder, obj):
        return self.default_decoder(obj)


class CBORSerializer(CustomizableSerializer):
    """
    Serializes objects using CBOR (Concise Binary Object Representation).

    To use this serializer backend, the ``cbor2`` library must be installed.
    A convenient way to do this is to install ``asphalt-serialization`` with the ``cbor`` extra:

    .. code-block:: shell

        $ pip install asphalt-serialization[cbor]

    .. seealso:: `cbor2 documentation <https://pypi.io/project/cbor2/>`_

    :param encoder_options: keyword arguments passed to ``cbor2.dumps()``
    :param decoder_options: keyword arguments passed to ``cbor2.loads()``
    :param custom_type_codec: wrapper to use to wrap custom types after marshalling, or ``None`` to
        return marshalled objects as-is
    """

    __slots__ = ('encoder_options', 'decoder_options', 'custom_type_codec', 'marshallers',
                 'unmarshallers')

    def __init__(self, encoder_options: Dict[str, Any] = None,
                 decoder_options: Dict[str, Any] = None,
                 custom_type_codec: Union[CBORTypeCodec, str] = None) -> None:
        assert check_argument_types()
        super().__init__(resolve_reference(custom_type_codec) or CBORTypeCodec())
        self.encoder_options = encoder_options or {}
        self.decoder_options = decoder_options or {}

    def serialize(self, obj) -> bytes:
        return cbor2.dumps(obj, **self.encoder_options)

    def deserialize(self, payload: bytes):
        return cbor2.loads(payload, **self.decoder_options)

    @property
    def mimetype(self):
        return 'application/cbor'



from __future__ import annotations

import pickle

from typeguard import check_argument_types

from asphalt.serialization.api import Serializer


class PickleSerializer(Serializer):
    """
    Serializes objects using the standard library :mod:`pickle` module.

    .. warning:: This serializer is insecure because it allows execution of arbitrary code when
        deserializing. Avoid using this if at all possible.

    :param protocol: pickle protocol level to use (defaults to the highest possible)
    """

    __slots__ = 'protocol'

    def __init__(self, protocol: int = pickle.HIGHEST_PROTOCOL):
        assert check_argument_types()
        assert 0 <= protocol <= pickle.HIGHEST_PROTOCOL,\
            f'"protocol" must be between 0 and {pickle.HIGHEST_PROTOCOL}'

        self.protocol = protocol

    def serialize(self, obj) -> bytes:
        return pickle.dumps(obj, protocol=self.protocol)

    def deserialize(self, payload: bytes):
        return pickle.loads(payload)

    @property
    def mimetype(self):
        return 'application/python-pickle'



import re
from datetime import datetime, timezone
from functools import partial
from types import SimpleNamespace

import pytest
from cbor2 import CBORTag
from msgpack import ExtType

from asphalt.serialization.serializers.cbor import CBORSerializer, CBORTypeCodec
from asphalt.serialization.serializers.json import JSONSerializer
from asphalt.serialization.serializers.msgpack import MsgpackSerializer, MsgpackTypeCodec
from asphalt.serialization.serializers.pickle import PickleSerializer
from asphalt.serialization.serializers.yaml import YAMLSerializer


class SimpleType:
    def __init__(self, value_a, value_b):
        self.value_a = value_a
        self.value_b = value_b

    def __eq__(self, other):
        if isinstance(other, SimpleType):
            return other.value_a == self.value_a and other.value_b == self.value_b
        return NotImplemented


class SlottedSimpleType:
    __slots__ = 'value_a', 'value_b'

    def __init__(self, value_a, value_b):
        self.value_a = value_a
        self.value_b = value_b

    def __getstate__(self):
        return {'value_a': self.value_a, 'value_b': self.value_b}

    def __setstate__(self, state):
        self.value_a = state['value_a']
        self.value_b = state['value_b']

    def __eq__(self, other):
        if isinstance(other, SlottedSimpleType):
            return other.value_a == self.value_a and other.value_b == self.value_b
        return NotImplemented


class CustomStateSimpleType(SimpleType):
    def unmarshal(self, state):
        self.value_a, self.value_b = state

    def marshal(self):
        return self.value_a, self.value_b


class UnserializableSimpleType:
    __slots__ = 'value_a', 'value_b'

    def __init__(self, value_a, value_b):
        self.value_a = value_a
        self.value_b = value_b


def marshal_datetime(dt: datetime):
    return dt.timestamp()


def unmarshal_datetime(state):
    return datetime.fromtimestamp(state, timezone.utc)


@pytest.fixture(params=['cbor', 'json', 'msgpack', 'pickle', 'yaml'])
def serializer_type(request):
    return request.param


@pytest.fixture
def serializer(request, serializer_type):
    kwargs = getattr(request, 'param', {})
    return {
        'cbor': partial(CBORSerializer, encoder_options=dict(value_sharing=True)),
        'json': JSONSerializer,
        'msgpack': MsgpackSerializer,
        'pickle': PickleSerializer,
        'yaml': YAMLSerializer
    }[serializer_type](**kwargs)


@pytest.mark.parametrize('input', [
    'åäö',
    -8,
    5.06,
    [1, 'test', 1.23],
    {'x': 'foo', 'bar': 'baz'}
], ids=['str', 'int', 'float', 'list', 'dict'])
def test_basic_types_roundtrip(serializer, input):
    output = serializer.serialize(input)
    assert isinstance(output, bytes)

    deserialized = serializer.deserialize(output)
    assert deserialized == input


@pytest.mark.parametrize('serializer_type', ['cbor', 'pickle', 'yaml'])
def test_circular_reference(serializer):
    a = {'foo': 1}
    b = {'a': a}
    a['b'] = b

    output = serializer.serialize(a)
    assert isinstance(output, bytes)

    other_a = serializer.deserialize(output)
    assert other_a['foo'] == 1
    other_b = other_a['b']
    assert other_b['a'] is other_a


@pytest.mark.parametrize('serializer_type', ['cbor', 'msgpack', 'json'])
class TestCustomTypes:
    @pytest.mark.parametrize('cls', [SimpleType, SlottedSimpleType], ids=['normal', 'slotted'])
    def test_marshal_unmarshal(self, serializer, cls):
        serializer.register_custom_type(cls)
        testval = cls(1, {'a': 1})
        testval2 = cls(2, testval)
        output = serializer.serialize(testval2)
        outval = serializer.deserialize(output)
        assert outval == testval2

    def test_custom_state(self, serializer):
        """Test that marshallers and umarshallers can be embedded into the relevant class."""
        serializer.register_custom_type(CustomStateSimpleType, CustomStateSimpleType.marshal,
                                        CustomStateSimpleType.unmarshal)
        testval = CustomStateSimpleType('a', 1)
        output = serializer.serialize(testval)
        outval = serializer.deserialize(output)
        assert outval == testval

    def test_marshal_builtin(self, serializer):
        """Test that a single-argument unmarshaller can be used to unmarshal built-in types."""
        serializer.register_custom_type(datetime, marshal_datetime, unmarshal_datetime)
        dt = datetime(2016, 9, 9, 7, 21, 16, tzinfo=timezone.utc)
        output = serializer.serialize(dt)
        dt2 = serializer.deserialize(output)
        assert dt == dt2

    def test_missing_getattr(self, serializer):
        testval = UnserializableSimpleType(1, 'a')
        serializer.register_custom_type(UnserializableSimpleType)
        exc = pytest.raises(TypeError, serializer.serialize, testval)
        exc.match("'test_serializers.UnserializableSimpleType' has no __dict__ attribute and does "
                  "not implement __getstate__()")

    def test_missing_setattr(self, serializer):
        testval = UnserializableSimpleType(1, 'a')
        serializer.register_custom_type(UnserializableSimpleType, lambda instance: {})
        serialized = serializer.serialize(testval)
        exc = pytest.raises(Exception, serializer.deserialize, serialized)
        exc.match(
            "'test_serializers.UnserializableSimpleType' has no __dict__ attribute and does not "
            "implement __setstate__()")

    def test_missing_marshaller(self, serializer_type, serializer):
        serializer.register_custom_type(SlottedSimpleType)
        testval = SimpleType(1, 'a')
        exc = pytest.raises(Exception, serializer.serialize, testval)
        exc.match('no marshaller found for type "test_serializers.SimpleType"')

    def test_missing_unmarshaller(self, serializer):
        serializer.register_custom_type(SlottedSimpleType)
        serializer.register_custom_type(SimpleType, unmarshaller=None)
        testval = SimpleType(1, 'a')
        serialized = serializer.serialize(testval)
        exc = pytest.raises(Exception, serializer.deserialize, serialized)
        exc.match('no unmarshaller found for type "test_serializers.SimpleType"')

    def test_nowrap(self, serializer):
        serializer.register_custom_type(SimpleType, wrap_state=False)
        testval = SimpleType(1, 'a')
        serialized = serializer.serialize(testval)
        deserialized = serializer.deserialize(serialized)
        assert deserialized == {'value_a': 1, 'value_b': 'a'}


def test_mime_types(serializer):
    assert re.match('[a-z]+/[a-z]+', serializer.mimetype)


@pytest.mark.parametrize('safe', [True, False], ids=['safe', 'unsafe'])
def test_yaml_safe_attribute(safe):
    serializer = YAMLSerializer(safe=safe)
    assert serializer.safe is safe


@pytest.mark.parametrize('serializer_type', ['msgpack'])
def test_msgpack_exttype_passthrough(serializer):
    serializer.register_custom_type(SlottedSimpleType)
    ext = ExtType(6, b'somedata')
    data = serializer.serialize(ext)
    obj = serializer.deserialize(data)
    assert isinstance(obj, ExtType)
    assert obj.code == 6
    assert obj.data == b'somedata'


@pytest.mark.parametrize('serializer_type', ['cbor'])
def test_cbor_self_referential_objects(serializer):
    value1 = SimpleNamespace()
    value1.val = 1
    value1.next = value2 = SimpleNamespace()
    value2.val = 2
    value2.previous = value1

    serializer.register_custom_type(SimpleNamespace, typename='Simple')
    data = serializer.serialize(value1)
    obj = serializer.deserialize(data)
    assert obj.val == 1
    assert obj.next.val == 2
    assert obj.next.previous is obj


@pytest.mark.parametrize('serializer_type', ['cbor'])
def test_cbor_oneshot_unmarshal(serializer):
    def unmarshal_simple(state):
        return SimpleType(**state)

    obj = SimpleType(1, 2)
    serializer.register_custom_type(SimpleType, unmarshaller=unmarshal_simple)
    data = serializer.serialize(obj)
    obj = serializer.deserialize(data)
    assert obj.value_a == 1
    assert obj.value_b == 2


@pytest.mark.parametrize('serializer_type', ['cbor'])
def test_cbor_raw_tag(serializer):
    tag = CBORTag(6000, 'Hello')
    serializer.register_custom_type(SimpleType)
    data = serializer.serialize(tag)
    tag = serializer.deserialize(data)
    assert tag.tag == 6000
    assert tag.value == 'Hello'


class TestObjectHook:
    @pytest.fixture(params=['msgpack', 'cbor'])
    def serializer(self, request):
        if request.param == 'msgpack':
            codec = MsgpackTypeCodec(type_code=None)
            return MsgpackSerializer(custom_type_codec=codec)
        else:
            codec = CBORTypeCodec(type_tag=None)
            return CBORSerializer(custom_type_codec=codec)

    def test_object_hook(self, serializer):
        value1 = SimpleNamespace()
        value1.val = 1
        value1.next = value2 = SimpleNamespace()
        value2.val = 2

        serializer.register_custom_type(SimpleNamespace, typename='Simple')
        data = serializer.serialize(value1)
        obj = serializer.deserialize(data)
        assert obj.val == 1
        assert obj.next.val == 2



"""A simple example that serializes a dictionary and prints out the result."""
import asyncio

from asphalt.core import ContainerComponent, Context, run_application


class ApplicationComponent(ContainerComponent):
    async def start(self, ctx: Context):
        self.add_component('serialization', backend='json')
        await super().start(ctx)

        payload = ctx.json.serialize(
            {'a': 1, 'b': 5.03, 'c': [1, 2, 3], 'd': {'x': 'nested'}})
        print('JSON serialized dict:', payload.decode())
        asyncio.get_event_loop().stop()

run_application(ApplicationComponent())



from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Tuple

from asphalt.core import Component, Context, PluginContainer, merge_config
from typeguard import check_argument_types

from asphalt.serialization.api import CustomizableSerializer, Serializer

serializer_types = PluginContainer('asphalt.serialization.serializers', Serializer)
logger = logging.getLogger(__name__)


class SerializationComponent(Component):
    """
    Creates one or more serializer resources.

    Every serializer resource will be available in the context as the following types:

    * :class:`~asphalt.serialization.api.Serializer`
    * :class:`~asphalt.serialization.api.CustomizableSerializer` (if the serializer implements it)
    * its actual type

    Serializers can be configured in two ways:

    #. a single serializer, with configuration supplied directly as keyword arguments to this
        component's constructor (with the resource name being ``default`` and the context attribute
        matching the backend name)
    #. multiple serializers, by providing the ``serializers`` option where each key is the resource
        name and each value is a dictionary containing that serializer's configuration (with the
        context attribute matching the resource name by default)

    Each serializer configuration has two special options that are not passed to the constructor of
    the backend class:

    * backend: entry point name of the serializer backend class (required)
    * context_attr: name of the context attribute of the serializer resource

    :param serializers: a dictionary of resource name ⭢ constructor arguments for the chosen
        backend class
    :param default_serializer_args: default values for constructor keyword arguments
    """

    def __init__(self, serializers: Dict[str, Optional[Dict[str, Any]]] = None,
                 **default_serializer_args):
        assert check_argument_types()
        if not serializers:
            default_serializer_args.setdefault(
                'context_attr', default_serializer_args.get('backend'))
            serializers = {'default': default_serializer_args}

        self.serializers: List[Tuple] = []
        for resource_name, config in serializers.items():
            merged = merge_config(default_serializer_args, config or {})
            type_ = merged.pop('backend', resource_name)
            context_attr = merged.pop('context_attr', resource_name)
            serializer = serializer_types.create_object(type_, **merged)
            self.serializers.append((resource_name, context_attr, serializer))

    async def start(self, ctx: Context):
        for resource_name, context_attr, serializer in self.serializers:
            types = [Serializer, type(serializer)]
            if isinstance(serializer, CustomizableSerializer):
                types.append(CustomizableSerializer)

            ctx.add_resource(serializer, resource_name, context_attr, types=types)
            logger.info('Configured serializer (%s / ctx.%s; type=%s)', resource_name,
                        context_attr, serializer.mimetype)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from inspect import signature
from typing import Any, Callable, Dict, Optional, Tuple, Union

from typeguard import check_argument_types, qualified_name

from asphalt.serialization.marshalling import default_marshaller, default_unmarshaller


class Serializer(metaclass=ABCMeta):
    """
    This abstract class defines the serializer API.

    Each serializer is required to support the serialization of the following Python types,
    at minimum:

    * :class:`str`
    * :class:`int`
    * :class:`float`
    * :class:`list`
    * :class:`dict` (with ``str`` keys)

    A subclass may support a wider range of types, along with hooks to provide serialization
    support for custom types.
    """

    __slots__ = ()

    @abstractmethod
    def serialize(self, obj) -> bytes:
        """Serialize a Python object into bytes."""

    @abstractmethod
    def deserialize(self, payload: bytes):
        """Deserialize bytes into a Python object."""

    @property
    @abstractmethod
    def mimetype(self) -> str:
        """Return the MIME type for this serialization format."""


class CustomizableSerializer(Serializer):
    """
    This abstract class defines an interface for registering custom types on a serializer so that
    the the serializer can be extended to (de)serialize a broader array of classes.

    :ivar marshallers: a mapping of class -> (typename, marshaller callback)
    :vartype marshallers: Dict[str, Callable]
    :ivar unmarshallers: a mapping of class -> (typename, unmarshaller callback)
    :vartype unmarshallers: Dict[str, Callable]
    """

    __slots__ = ('custom_type_codec', 'marshallers', 'unmarshallers')

    def __init__(self, custom_type_codec: 'CustomTypeCodec'):
        self.custom_type_codec = custom_type_codec
        self.marshallers: Dict[type, Tuple[str, Callable, bool]] = {}
        self.unmarshallers: Dict[str, Tuple[Optional[type], Callable]] = {}

    def register_custom_type(
            self, cls: type, marshaller: Optional[Callable[[Any], Any]] = default_marshaller,
            unmarshaller: Union[Callable[[Any, Any], None],
                                Callable[[Any], Any], None] = default_unmarshaller, *,
            typename: str = None, wrap_state: bool = True) -> None:
        """
        Register a marshaller and/or unmarshaller for the given class.

        The state object returned by the marshaller and passed to the unmarshaller can be any
        serializable type. Usually a dictionary mapping of attribute names to values is used.

        .. warning:: Registering marshallers/unmarshallers for any custom type will override any
            serializer specific encoding/decoding hooks (respectively) already in place!

        :param cls: the class to register
        :param marshaller: a callable that takes the object to be marshalled as the argument and
              returns a state object
        :param unmarshaller: a callable that either:

            * takes an uninitialized instance of ``cls`` and its state object as arguments and
              restores the state of the object
            * takes a state object and returns a new instance of ``cls``
        :param typename: a unique identifier for the type (defaults to the ``module:varname``
            reference to the class)
        :param wrap_state: ``True`` to wrap the marshalled state before serialization so that it
            can be recognized later for unmarshalling, ``False`` to serialize it as is

        """
        assert check_argument_types()
        typename = typename or qualified_name(cls)

        if marshaller:
            self.marshallers[cls] = typename, marshaller, wrap_state
            self.custom_type_codec.register_object_encoder_hook(self)

        if unmarshaller and self.custom_type_codec is not None:
            target_cls: Optional[type] = cls
            if len(signature(unmarshaller).parameters) == 1:
                target_cls = None

            self.unmarshallers[typename] = target_cls, unmarshaller
            self.custom_type_codec.register_object_decoder_hook(self)


class CustomTypeCodec(metaclass=ABCMeta):
    """Interface for customizing how custom types are encoded and decoded."""

    @abstractmethod
    def register_object_encoder_hook(self, serializer: CustomizableSerializer) -> None:
        """
        Register a custom encoder callback on the serializer.

        This callback would be called when the serializer encounters an object it cannot natively
        serialize. What the callback returns is specific to each serializer type.

        :param serializer: the serializer instance to use
        """

    @abstractmethod
    def register_object_decoder_hook(self, serializer: CustomizableSerializer) -> None:
        """
        Register a callback on the serializer for unmarshalling previously marshalled objects.

        :param serializer: the serializer instance to use
        """



from __future__ import annotations

from typing import Any, Dict, Optional, Tuple, Union

from asphalt.core import resolve_reference
from msgpack import ExtType, packb, unpackb
from typeguard import check_argument_types

from asphalt.serialization.api import CustomizableSerializer
from asphalt.serialization.object_codec import DefaultCustomTypeCodec


class MsgpackTypeCodec(DefaultCustomTypeCodec):
    """
    Default custom type codec implementation for :class:`~.MsgpackSerializer`.

    Wraps marshalled state in either msgpack's ExtType objects (the default) or dicts (with
    ``type_code=None``).

    :param type_code: msgpack type code to use, or ``None`` to use JSON compatible dict-based
        wrapping
    """

    def __init__(self, type_code: Optional[int] = 119, **kwargs):
        assert check_argument_types()
        super().__init__(**kwargs)
        self.type_code = type_code

        if type_code:
            self.wrap_callback = self.wrap_state_ext_type
            self.unwrap_callback = self.unwrap_state_ext_type

    def register_object_decoder_hook(self, serializer: 'MsgpackSerializer') -> None:
        self.serializer = serializer
        serializer.packer_options['default'] = self.default_encoder

    def register_object_encoder_hook(self, serializer: 'MsgpackSerializer') -> None:
        self.serializer = serializer
        if self.type_code:
            serializer.unpacker_options['ext_hook'] = self.ext_hook
        else:
            serializer.unpacker_options['object_hook'] = self.default_decoder

    def ext_hook(self, code: int, data: bytes):
        if code == self.type_code:
            return self.default_decoder(data)
        else:
            return ExtType(code, data)

    def wrap_state_ext_type(self, typename: str, state):
        data = typename.encode('utf-8') + b':' + self.serializer.serialize(state)
        return ExtType(self.type_code, data)

    def unwrap_state_ext_type(
            self, wrapped_state: bytes) -> Union[Tuple[str, Any], Tuple[None, None]]:
        typename, payload = wrapped_state.split(b':', 1)
        return typename.decode('utf-8'), self.serializer.deserialize(payload)


class MsgpackSerializer(CustomizableSerializer):
    """
    Serializes objects using the msgpack library.

    The following defaults are passed to packer/unpacker and can be overridden by setting values
    for the options explicitly:

    * ``use_bin_type=True`` (packer)
    * ``raw=False`` (unpacker)

    To use this serializer backend, the ``msgpack`` library must be installed.
    A convenient way to do this is to install ``asphalt-serialization`` with the ``msgpack``
    extra:

    .. code-block:: shell

        $ pip install asphalt-serialization[msgpack]

    .. seealso:: `Msgpack web site <http://msgpack.org/>`_

    :param packer_options: keyword arguments passed to :func:`msgpack.packb`
    :param unpacker_options: keyword arguments passed to :func:`msgpack.unpackb`
    :param custom_type_codec: wrapper to use to wrap custom types after marshalling, or ``None`` to
        return marshalled objects as-is
    """

    __slots__ = ('packer_options', 'unpacker_options', 'custom_type_codec', '_marshallers',
                 '_unmarshallers')

    def __init__(
            self, packer_options: Dict[str, Any] = None, unpacker_options: Dict[str, Any] = None,
            custom_type_codec: Union[MsgpackTypeCodec, str] = None) -> None:
        assert check_argument_types()
        super().__init__(resolve_reference(custom_type_codec) or MsgpackTypeCodec())
        self.packer_options = packer_options or {}
        self.packer_options.setdefault('use_bin_type', True)
        self.unpacker_options = unpacker_options or {}
        self.unpacker_options.setdefault('raw', False)

    def serialize(self, obj) -> bytes:
        return packb(obj, **self.packer_options)

    def deserialize(self, payload: bytes):
        return unpackb(payload, **self.unpacker_options)

    @property
    def mimetype(self):
        return 'application/msgpack'



from __future__ import annotations

from typing import Any, Callable, Dict, Tuple, Union

from asphalt.core import qualified_name
from typeguard import check_argument_types

from asphalt.serialization.api import CustomizableSerializer, CustomTypeCodec


class DefaultCustomTypeCodec(CustomTypeCodec):
    """
    Provides default wrappers for implementing :class:`~asphalt.serialization.api.CustomTypeCodec`.

    :ivar CustomizableSerializer serializer: the associated serializer

    :param type_key: dict key for the type name
    :param state_key dict key for the marshalled state
    """

    serializer: CustomizableSerializer

    def __init__(self, type_key: str = '__type__', state_key: str = 'state'):
        assert check_argument_types()
        self.type_key = type_key
        self.state_key = state_key
        self.wrap_callback: Callable[[str, Any], Any] = self.wrap_state_dict
        self.unwrap_callback: Callable[[Any], Any] = self.unwrap_state_dict

    def default_encoder(self, obj):
        obj_type = obj.__class__
        try:
            typename, marshaller, wrap_state = self.serializer.marshallers[obj_type]
        except KeyError:
            raise LookupError('no marshaller found for type "{}"'
                              .format(qualified_name(obj_type))) from None

        state = marshaller(obj)
        return self.wrap_callback(typename, state) if wrap_state else state

    def default_decoder(self, obj):
        """Handle a dict that might contain a wrapped state for a custom type."""
        typename, marshalled_state = self.unwrap_callback(obj)
        if typename is None:
            return obj

        try:
            cls, unmarshaller = self.serializer.unmarshallers[typename]
        except KeyError:
            raise LookupError(f'no unmarshaller found for type "{typename}"') from None

        if cls is not None:
            instance = cls.__new__(cls)
            unmarshaller(instance, marshalled_state)
            return instance
        else:
            return unmarshaller(marshalled_state)

    def wrap_state_dict(self, typename: str, state) -> Dict[str, Any]:
        """
        Wrap the marshalled state in a dictionary.

        The returned dictionary has two keys, corresponding to the ``type_key`` and ``state_key``
        options. The former holds the type name and the latter holds the marshalled state.

        :param typename: registered name of the custom type
        :param state: the marshalled state of the object
        :return: an object serializable by the serializer

        """
        return {self.type_key: typename, self.state_key: state}

    def unwrap_state_dict(self, obj: Dict[str, Any]) -> Union[Tuple[str, Any], Tuple[None, None]]:
        """Unwraps a marshalled state previously wrapped using :meth:`wrap_state_dict`."""
        if len(obj) == 2:
            typename = obj.get(self.type_key)
            state = obj.get(self.state_key)
            if typename is not None:
                return typename, state

        return None, None



"""An example that demonstrates how to serialize custom types."""
import asyncio

from asphalt.core import ContainerComponent, Context, run_application


class Book:
    def __init__(self, name, author, year, isbn, sequel=None):
        self.name = name
        self.author = author
        self.year = year
        self.isbn = isbn
        self.sequel = sequel


class ApplicationComponent(ContainerComponent):
    async def start(self, ctx: Context):
        self.add_component('serialization', backend='json')
        await super().start(ctx)

        ctx.json.register_custom_type(Book, typename='Book')  # typename is optional
        book2 = Book('The Fall of Hyperion', 'Dan Simmons', 1995, '978-0553288209')
        book1 = Book('Hyperion', 'Dan Simmons', 1989, '978-0553283686', book2)
        payload = ctx.json.serialize(book1)
        print('JSON serialized dict:', payload.decode())
        asyncio.get_event_loop().stop()

run_application(ApplicationComponent())
